from powergenome.run_powergenome_multiple_outputs_cli import main

main(
    # settings_file="/Users/gs5183/Documents/debugging/MISO_622/settings",
    # settings_file="/Users/gs5183/Documents/PowerGenome/example_systems/CA_AZ/settings",
    # settings_file="/Users/gs5183/Documents/MIP/settings_conus_3z",
    # settings_file="/Users/gs5183/Documents/debugging/PJM/PJM_test_settings (3).yml",
    # settings_file="/Users/gs5183/Documents/debugging/MISO_802/settings",
    # settings_file="/Users/gs5183/Documents/MIP/MIP_results_comparison/case_settings/26-zone/settings",
    # settings_file="/Users/gs5183/Documents/debugging/Annie/conus_7z_settings.yml",
    # settings_file="/Users/gs5183/Documents/MIP/settings",
    # settings_file="/Users/gs5183/Documents/debugging/Gabe_conus/conus_26z_settings.yml",
    # settings_file="/Users/gs5183/Documents/debugging/PJM/settings",
    # settings_file="/Users/gs5183/Documents/MIP/MIP_results_comparison/case_settings/26-zone/settings-atb2023",
    # settings_file="/Users/gs5183/Documents/MIP/MIP_results_comparison/case_settings/26-zone/settings-atb2023",
    # results_folder="cluster_with_retired_check",
    # settings_file="/Users/gs5183/Documents/debugging/riti/WECC_EV_PG_runs/WECC_IRA_ALL_EV_6_shift.yml",
    # settings_file="/Users/gs5183/Documents/PG_setups/WECC_geothermal_dac/settings-atb2023",
    # settings_file="/Users/gs5183/Documents/PG_setups/PJM_Avery/pg_settings",
    # settings_file="/Users/gs5183/Documents/PG_setups/dac_ercot_settings/settings-15-zone",
    # settings_file="/Users/gs5183/Documents/MIP_results_comparison/case_settings/26-zone/settings-atb2023",
    # settings_file="/Users/gs5183/Documents/pg-stock-systems/CONUS_25z_inputs/settings",
    # settings_file="/Users/gs5183/Documents/researcher projects/Jacob/EI fission/Eastern/pg_settings_atb_2023",
    settings_file="/Users/gs5183/Documents/researcher projects/MN_CES/settings",
    # settings_file="/Users/gs5183/Documents/PowerGenome/example_systems/ISONE/settings",
    results_folder="test_0.4",
    # case_id=["t52nr"],
    # case_id=["current_policies_52_week"],
    # ["current_policies_52_week_profile"],  # ["base_short"],  # ["current_policies_52_week_profile"],
    # multi_period=True,
    current_gens=False,
    gens=True,
    # current_gens=False,
    load=True,
)

# main(
#     # settings_file="/Users/gs5183/Documents/debugging/MISO_622/settings",
#     settings_file="/Users/gs5183/Documents/PowerGenome/example_systems/CONUS-3-zone/settings",
#     # settings_file="/Users/gs5183/Documents/PowerGenome/example_systems/ISONE/issue_settings.yaml",
#     # settings_file="/Users/gs5183/Documents/MIP/settings_conus_3z",
#     # settings_file="/Users/gs5183/Documents/debugging/PJM/PJM_test_settings (3).yml",
#     # settings_file="/Users/gs5183/Documents/debugging/MISO_802/settings",
#     # settings_file="/Users/gs5183/Documents/MIP/MIP_results_comparison/case_settings/26-zone/settings",
#     # settings_file="/Users/gs5183/Documents/debugging/Annie/conus_7z_settings.yml",
#     # settings_file="/Users/gs5183/Documents/MIP/settings",
#     # settings_file="/Users/gs5183/Documents/debugging/Gabe_conus/conus_26z_settings.yml",
#     # settings_file="/Users/gs5183/Documents/debugging/PJM/settings",
#     # settings_file="/Users/gs5183/Documents/MIP/MIP_results_comparison/case_settings/26-zone/settings-atb2023",
#     # settings_file="/Users/gs5183/Documents/debugging/Fangwei/pg inputs/settings",
#     # settings_file="/Users/gs5183/Documents/PowerGenome/example_systems/ISONE/settings",
#     # results_folder="cluster_with_retired_check",
#     # settings_file="/Users/gs5183/Documents/debugging/riti/WECC_EV_PG_runs/WECC_IRA_ALL_EV_6_shift.yml",
#     # settings_file="/Users/gs5183/Documents/PG_setups/WECC_geothermal_dac/settings-atb2023",
#     # settings_file="/Users/gs5183/Documents/PG_setups/PJM_Avery/pg_settings",
#     results_folder="debug",
#     # case_id=["short_test_vre_buildout"],
#     # ["current_policies_52_week_profile"],  # ["base_short"],  # ["current_policies_52_week_profile"],
#     multi_period=True,
#     current_gens=True,
#     # current_gens=False,
#     load=False,
# )
